@@ -353,7 +353,8 @@ _wnck_window_create (Window      xwindow,
   window->priv->pid =
     _wnck_get_pid (window->priv->xwindow);
 
-  _wnck_get_window_geometry (xwindow,
+  _wnck_get_window_geometry (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			     xwindow,
                              &window->priv->x,
                              &window->priv->y,
                              &window->priv->width,
@@ -631,7 +632,8 @@ wnck_window_close (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_close (window->priv->xwindow);
+  _wnck_close (WNCK_SCREEN_XSCREEN (window->priv->screen),
+	       window->priv->xwindow);
 }
 
 void
@@ -655,7 +657,8 @@ wnck_window_maximize                (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"));
@@ -666,7 +669,8 @@ wnck_window_unmaximize              (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"));
@@ -677,7 +681,8 @@ wnck_window_maximize_horizontally   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"),
                       0);
@@ -688,7 +693,8 @@ wnck_window_unmaximize_horizontally (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_HORZ"),
                       0);
@@ -699,7 +705,8 @@ wnck_window_maximize_vertically     (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       0);
@@ -710,7 +717,8 @@ wnck_window_unmaximize_vertically   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_MAXIMIZED_VERT"),
                       0);
@@ -721,7 +729,8 @@ wnck_window_shade                   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_SHADED"),
                       0);
@@ -732,7 +741,8 @@ wnck_window_unshade                 (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_SHADED"),
                       0);
@@ -743,7 +753,8 @@ wnck_window_stick                   (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       TRUE,
                       _wnck_atom_get ("_NET_WM_STATE_STICKY"),
                       0);
@@ -754,7 +765,8 @@ wnck_window_unstick                 (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_state (window->priv->xwindow,
+  _wnck_change_state (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		      window->priv->xwindow,
                       FALSE,
                       _wnck_atom_get ("_NET_WM_STATE_STICKY"),
                       0);
@@ -776,7 +788,7 @@ wnck_window_get_workspace (WnckWindow *window)
   if (window->priv->workspace == ALL_WORKSPACES)
     return NULL;
   else
-    return wnck_workspace_get (window->priv->workspace);
+    return wnck_screen_get_workspace (window->priv->screen, window->priv->workspace);
 }
 
 void
@@ -786,7 +798,8 @@ wnck_window_move_to_workspace (WnckWindow    *window,
   g_return_if_fail (WNCK_IS_WINDOW (window));
   g_return_if_fail (WNCK_IS_WORKSPACE (space));
 
-  _wnck_change_workspace (window->priv->xwindow,
+  _wnck_change_workspace (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			  window->priv->xwindow,
                           wnck_workspace_get_number (space));
 }
 
@@ -813,7 +826,8 @@ wnck_window_pin (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_change_workspace (window->priv->xwindow,
+  _wnck_change_workspace (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			  window->priv->xwindow,
                           ALL_WORKSPACES);
 }
 
@@ -840,7 +854,8 @@ wnck_window_unpin (WnckWindow *window)
 
   active = wnck_screen_get_active_workspace (window->priv->screen);
 
-  _wnck_change_workspace (window->priv->xwindow,
+  _wnck_change_workspace (WNCK_SCREEN_XSCREEN (window->priv->screen),
+			  window->priv->xwindow,
                           active ? wnck_workspace_get_number (active) : 0);
 }
 
@@ -856,7 +871,8 @@ wnck_window_activate (WnckWindow *window)
 {
   g_return_if_fail (WNCK_IS_WINDOW (window));
 
-  _wnck_activate (window->priv->xwindow);
+  _wnck_activate (WNCK_SCREEN_XSCREEN (window->priv->screen),
+		  window->priv->xwindow);
 }
 
 /**
@@ -1168,7 +1184,7 @@ _wnck_window_process_property_notify (WnckWindow *window,
            xevent->xproperty.atom ==
            _wnck_atom_get ("KWM_WIN_ICON") ||
            xevent->xproperty.atom ==
-           _wnck_atom_get ("WM_NORMAL_HINTS"))
+           _wnck_atom_get ("WM_HINTS"))
     {
       _wnck_icon_cache_property_changed (window->priv->icon_cache,
                                          xevent->xproperty.atom);
@@ -1187,7 +1203,8 @@ _wnck_window_process_configure_notify (WnckWindow *window,
     }
   else
     {
-      _wnck_get_window_position (window->priv->xwindow,
+      _wnck_get_window_position (WNCK_SCREEN_XSCREEN (window->priv->screen),
+				 window->priv->xwindow,
                                  &window->priv->x,
                                  &window->priv->y);
     }
@@ -1283,12 +1300,12 @@ update_state (WnckWindow *window)
     case WNCK_WINDOW_MENU:
     case WNCK_WINDOW_UTILITY:
     case WNCK_WINDOW_SPLASHSCREEN:
+    case WNCK_WINDOW_DIALOG:
+    case WNCK_WINDOW_MODAL_DIALOG:
       window->priv->skip_taskbar = TRUE;
       break;
       
     case WNCK_WINDOW_NORMAL:
-    case WNCK_WINDOW_DIALOG:
-    case WNCK_WINDOW_MODAL_DIALOG:
       break;
     }
 
@@ -1389,29 +1406,90 @@ update_workspace (WnckWindow *window)
 static void
 update_actions (WnckWindow *window)
 {
+  Atom *atoms;
+  int   n_atoms;
+  int   i;
+
   if (!window->priv->need_update_actions)
     return;
-  
+
   window->priv->need_update_actions = FALSE;
-  
-  /* FIXME read property */
-  window->priv->actions =
-    WNCK_WINDOW_ACTION_MOVE                    |
-    WNCK_WINDOW_ACTION_RESIZE                  |
-    WNCK_WINDOW_ACTION_SHADE                   |
-    WNCK_WINDOW_ACTION_STICK                   |
-    WNCK_WINDOW_ACTION_MAXIMIZE_HORIZONTALLY   |
-    WNCK_WINDOW_ACTION_MAXIMIZE_VERTICALLY     |
-    WNCK_WINDOW_ACTION_CHANGE_WORKSPACE        |
-    WNCK_WINDOW_ACTION_CLOSE                   |
-    WNCK_WINDOW_ACTION_UNMAXIMIZE_HORIZONTALLY |
-    WNCK_WINDOW_ACTION_UNMAXIMIZE_VERTICALLY   |
-    WNCK_WINDOW_ACTION_UNSHADE                 |
-    WNCK_WINDOW_ACTION_UNSTICK                 |
-    WNCK_WINDOW_ACTION_MAXIMIZE                |
-    WNCK_WINDOW_ACTION_UNMAXIMIZE              |
-    WNCK_WINDOW_ACTION_MINIMIZE                |
-    WNCK_WINDOW_ACTION_UNMINIMIZE;
+
+  window->priv->actions = 0;
+
+  if (!_wnck_get_atom_list (window->priv->xwindow,
+                            _wnck_atom_get ("_NET_WM_ALLOWED_ACTIONS"),
+                            &atoms,
+                            &n_atoms))
+    {
+      window->priv->actions = 
+                WNCK_WINDOW_ACTION_MOVE                    |
+                WNCK_WINDOW_ACTION_RESIZE                  |
+                WNCK_WINDOW_ACTION_SHADE                   |
+                WNCK_WINDOW_ACTION_STICK                   |
+                WNCK_WINDOW_ACTION_MAXIMIZE_HORIZONTALLY   |
+                WNCK_WINDOW_ACTION_MAXIMIZE_VERTICALLY     |
+                WNCK_WINDOW_ACTION_CHANGE_WORKSPACE        |
+                WNCK_WINDOW_ACTION_CLOSE                   |
+                WNCK_WINDOW_ACTION_UNMAXIMIZE_HORIZONTALLY |
+                WNCK_WINDOW_ACTION_UNMAXIMIZE_VERTICALLY   |
+                WNCK_WINDOW_ACTION_UNSHADE                 |
+                WNCK_WINDOW_ACTION_UNSTICK                 |
+                WNCK_WINDOW_ACTION_MAXIMIZE                |
+                WNCK_WINDOW_ACTION_UNMAXIMIZE              |
+                WNCK_WINDOW_ACTION_MINIMIZE                |
+                WNCK_WINDOW_ACTION_UNMINIMIZE;
+      return;
+    }
+
+  i = 0;
+  while (i < n_atoms)
+    {
+      if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_MOVE"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_MOVE;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_RESIZE"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_RESIZE;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_SHADE"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_SHADE |
+                                 WNCK_WINDOW_ACTION_UNSHADE;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_STICK"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_STICK |
+                                 WNCK_WINDOW_ACTION_UNSTICK;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_MAXIMIZE_HORZ"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_MAXIMIZE_HORIZONTALLY |
+                                 WNCK_WINDOW_ACTION_UNMAXIMIZE_HORIZONTALLY;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_MAXIMIZE_VERT"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_MAXIMIZE_VERTICALLY |
+                                 WNCK_WINDOW_ACTION_UNMAXIMIZE_VERTICALLY;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_CHANGE_DESKTOP"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_CHANGE_WORKSPACE;
+
+      else if (atoms[i] == _wnck_atom_get ("_NET_WM_ACTION_CLOSE"))
+        window->priv->actions |= WNCK_WINDOW_ACTION_CLOSE;
+      else
+        g_warning ("Unhandled action type %s", _wnck_atom_name (atoms [i]));
+
+      i++;
+    }
+
+  g_free (atoms);
+
+  if ((window->priv->actions & WNCK_WINDOW_ACTION_MAXIMIZE_HORIZONTALLY) &&
+      (window->priv->actions & WNCK_WINDOW_ACTION_MAXIMIZE_VERTICALLY))
+    window->priv->actions |=
+        WNCK_WINDOW_ACTION_MAXIMIZE   |
+        WNCK_WINDOW_ACTION_UNMAXIMIZE;
+
+  /* These are always enabled */
+  window->priv->actions |=
+        WNCK_WINDOW_ACTION_MINIMIZE   |
+        WNCK_WINDOW_ACTION_UNMINIMIZE;
 }
 
 static void
